/*****************************************************************************\
|*
|*  Version:    1.3
|*
|*  Description:
|*
|*      TASKING M16C ROM Monitor: command processor
|*
|*  Copyright 2003 Altium BV
|*
 *****************************************************************************/

#include <stdlib.h>
#include "target.h"
#include "uart.h"
#include "command.h"
#include "m16c.h"
#include "flash.h"
#include <setjmp.h>

/*
 * local function prototypes
 */

/* unsigned char cmd_conf(unsigned char c); is defined in command.h */
void cmd_reset( void);
void cmd_conf( void);
void cmd_debug( void);
void cmd_clr( void);
void cmd_dwnh( void);
void cmd_pb( void);
void cmd_pbcl( void);
void cmd_gb( void);
void cmd_prog( void);
void cmd_stop( void);
void cmd_step( void);
void cmd_xreg( void);
void cmd_dump( void);
void cmd_fill( void);
void cmd_fm( void);
void cmd_err( void);

char clear_all_bp(void);

/* function table, associate a function name (code) with a function to call */
__far __rom FUNC_MAP_T func_table[] =
{
    { CMD_CONF, cmd_conf },
    { CMD_DEBUG, cmd_debug },
    { CMD_REST, cmd_reset},
    { CMD_CLR,  cmd_clr  },
    { CMD_DWNH, cmd_dwnh },
    { CMD_PB,   cmd_pb   },
    { CMD_PBCL, cmd_pbcl },
    { CMD_GB,   cmd_gb   },
    { CMD_PROG, cmd_prog },
    { CMD_STOP, cmd_stop },
    { CMD_STEP, cmd_step },
    { CMD_XREG, cmd_xreg },
    { CMD_DUMP, cmd_dump },
    { CMD_FILL, cmd_fill },
    { CMD_FM,   cmd_fm   },
    { CMD_UNKNOWN, NULL  }
};

/**********************************************************************\
|*
|*  FUNCTION: cmd_conv_nibble
|*
|*  PARAMETERS: ASCII character containing nibble to convert
|*
|*  DESCRIPTION:
|*              Convert the nibble to the correct value.
|*      If a '(' is detected it is assumed a new command
|*              command is being started, so longjmp directly to
|*              the command handler.
|*
|*  RETURNS: the converted value, or CTRL_END if recvd
|*
 */
inline unsigned char cmd_conv_nibble( unsigned char c )
{
    if ( (c>='0') && (c<='9') )
    {
        return c - '0';
    }
    else if ( ((c|0x20)>='a') && ((c|0x20)<='f') )
    {
        return (c|0x20) - 'a' + 10;
    }
    else if ( c == ')' )
    {
            return c;
    }
    else if ( c == '(' )
    {
        longjmp( env, 1 );
    }
    else
    {
        err_code[1] = '8';
    }
    return SYNTAX_ERROR;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_getnibble
|*
|*  DESCRIPTION:
|*      While no syntax errors occured get one character
|*      from the serial port and convert it to a hex nibble
|*
|*  RETURNS: the converted value, or CTRL_END if one recvd
|*      on syntax error return 0.
|*
 */
static unsigned char
cmd_getnibble( void )
{
    unsigned char ch = 0;
    if ( err_code[1] == ERR_NORMAL )
    {
        ch = cmd_conv_nibble( uart_getchar() );
    }
    return ch;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_getendchar
|*
|*  DESCRIPTION:
|*      Check if the following character is an end character.
|*      If not mark this command containing a syntax error.
|*
 */
void
cmd_getendchar( void )
{
    if ( uart_getchar() != CTRL_END )
    {
        err_code[1] = SYNTAX_ERROR;
    }
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_getval56
|*
|*  Parameters:         size value marking to get 5 or 6 nibbles
|*
|*  DESCRIPTION:
|*      Get a 5 or 6 nibble value from the serial port.
|*      If the first character is an end character return it,
|*
 */
unsigned char
cmd_getval56( unsigned char size )
{
    unsigned char i;
    unsigned char ch;

    ch = cmd_getnibble();
    if ( ch == CTRL_END )
    {
        return CTRL_END;
    }
    val56 = ch;
    for ( i = 1; i < size; ++i )
    {
        val56 = (val56 << 4) | cmd_getnibble();
    }
    return 0;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_getval4
|*
|*  DESCRIPTION:
|*      Get a 4 nibble value from the serial port.
|*      If the first character is an end character return it,
|*
 */
unsigned char
cmd_getval4( void )
{
    unsigned char ch;
    unsigned char i;

    ch = cmd_getnibble();
    if ( ch == CTRL_END )
    {
        return CTRL_END;
    }
    val4 = ch;
    for ( i = 0; i < 3; ++i )
    {
        val4 = (val4 << 4) | cmd_getnibble();
    }
    return 0;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_getval2
|*
|*  DESCRIPTION:
|*      Get a 2 nibble value from the serial port.
|*      If the first character is an end character return it,
|*
 */
unsigned char
cmd_getval2( void )
{
    unsigned char ch;

    ch = cmd_getnibble();
    if ( ch == CTRL_END )
    {
        return CTRL_END;
    }
    val2 = ch << 4;
    ch = cmd_getnibble();
    val2 |= ch;
    return 0;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_set_handle
|*
|*  PARAMETERS: new_cmd_code = command code to set
|*
|*  DESCRIPTION: map a command number to function pointer index
|*
|*  RETURNS: nothing
|*
 */
void cmd_set_handle(unsigned char new_cmd_code)
{
    int i;

    /* find the function that matches the new command... */
    i=0;
    cmd_handle = cmd_err;
    while ( func_table[i].func_name != CMD_UNKNOWN )
    {
        if ( func_table[i].func_name == new_cmd_code )
        {
            cmd_handle = func_table[i].func_ptr;
            break;
        }
        i++;
    } 
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_show_conf
|*
|*  DESCRIPTION: send "Monitor Program Information Report"
|*
 */
void
cmd_show_conf( void )
{
    /* show info */
    uart_puts(err_code);
    ta_show_conf();
    uart_putchar( ')' );
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_conf
|*
|*  DESCRIPTION: answer "Monitor Program Information Report"
|*
 */
void cmd_conf( void )
{
    /* incoming byte should be CTRL_END, if not, set err_code */
    cmd_getendchar();
    cmd_show_conf();
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_reset
|*
|*  PARAMETERS: c, incoming byte for this command
|*
|*  DESCRIPTION: reset the MCU (only when user program is not running)
|*
|*  RETURNS: STATE_RESET
|*
 */
void cmd_reset( void )
{
    /* incoming byte should be CTRL_END */
    if ( uart_getchar() == CTRL_END)
    {
        /* check if user program is running */
        if ( application_running )
        {
            err_code[1] = ERR_EXEC;
        }
        else
        {
            ta_reset();
        }
    }
    else
    {
        err_code[1] = SYNTAX_ERROR;
    }
    /* show info */
    uart_puts(err_code);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_clr
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: initialisation of monitor program
|*
|*  RETURNS: STATE_RESET
|*
 */
void cmd_clr( void )
{
    /* incoming byte should be CTRL_END */
    if ( uart_getchar() == CTRL_END)
    {
        /* check if user program is running */
        if ( application_running )
        {
            err_code[1] = ERR_EXEC;
        }
        else
        {
            /* init: clear breakpoints, let XVW init registers */
            clear_all_bp();
        }
    }
    else
    {
        err_code[1] = SYNTAX_ERROR;
    }
    /* show info */
    uart_puts(err_code);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_dwnh
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: perform hex download to flash or erase:
|*               command (CMD_DWNH): erase flash
|*               command (CMD_DWNH 
|*                       n-times: (address count data chk)
|*                       CTRL_RCVE): write flash
|*
|*  RETURNS: STATE_RESET (command finished or error occured)
|*           STATE_CMD (so far, so good...)
 */
void cmd_dwnh( void )
{
    /* check if user program is executing */
    if ( application_running )
    {
        uart_puts("(4)");
        return;
    }
    if ( cmd_getval56(6) == CTRL_END )
    {
        m16c_erase_flash();
    }
    else
    {
        unsigned int i;
        unsigned char x;
        unsigned char checksum;

        cmd_getval4();
        checksum = 0;
        __sstrw( 0xffff, (__near int *)downloadbuf, sizeof(downloadbuf)/2 );
        downloadptr = downloadbuf + (val56 & 0xff);
        for ( i = 0; i < val4; ++i )
        {
            x = uart_getchar();
            checksum -= x;
            if ( x == CTRL_ESC )
            {
                x = uart_getchar();
                checksum -= x;
                x -= 0x40;
            }
            *downloadptr = x;
            ++downloadptr;
        }
        cmd_getval2();
        if ( (uart_getchar() != '.') || (uart_getchar() != ')') )
        {
            err_code[1] = SYNTAX_ERROR;
        }
        else
        {
            if ( checksum == val2 )
            {
                val56 &= ~0xffL;
                if ( val56 < ROM_MONITOR_START )
                {
                    m16c_write_flash( val56, downloadbuf );
                }
            }
            else
            {
                err_code[1] = ERR_CHECKSUM;
            }
        }
    }
    uart_puts(err_code);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_pb
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: set or clear a single breakpoint
|*               depending on global byte_count, the incoming byte is
|*               0 : set/clear switch
|*               1 : BP address byte1
|*               2 : BP address byte2
|*               3 : BP address byte3
|*               4 : BP address byte4
|*               5 : BP address byte5
|*               6 : should be CTRL_END
|*
|*  RETURNS: STATE_RESET (command finished or error occured)
|*           STATE_CMD (so far, so good...)
|*
 */
void cmd_pb( void )
{
    unsigned char set_clear = 0;

    /* get switch (0/1), byte_count == 0 */
    set_clear = uart_getchar();
    cmd_getval56(5);
    /* sanity check */
    if ( uart_getchar() == CTRL_END )
    {
        if ( set_clear == '1' )
        {
            /* clear breakpoint if available */
            if (break_points[0] == val56)
            {
                break_points[0] = 0x80000000;
            }
            else if (break_points[1] == val56)
            {
                break_points[1] = 0x80000000;
            }
#ifdef HAS_4BREAKPOINTS
            else if (break_points[2] == val56)
            {
                break_points[2] = 0x80000000;
            }
            else if (break_points[3] == val56)
            {
                break_points[3] = 0x80000000;
            }
#endif /* HAS_4BREAKPOINTS */
            else /* not found, set err_notbreak */
            {
                err_code[1] = ERR_NOTBREAK;
            }
        }
        else
        {
            /* set breakpoint if room available (err_code always ERR_NORMAL !)*/
            if (break_points[0] & 0x80000000)
            {
                break_points[0] = val56;
            }
            else if (break_points[1] & 0x80000000)
            {
                break_points[1] = val56;
            }
#ifdef HAS_4BREAKPOINTS
            else if (break_points[2] & 0x80000000)
            {
                break_points[2] = val56;
            }
            else if (break_points[3] & 0x80000000)
            {
                break_points[3] = val56;
            }
#endif /* HAS_4BREAKPOINTS */
        }
    }
    uart_puts(err_code);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_pbcl
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: clear breakpoints
|*
|*  RETURNS: STATE_RESET
|*
 */
void cmd_pbcl( void )
{
    /* incoming byte should be CTRL_END */
    if ( uart_getchar() == CTRL_END )
    {
        err_code[1] = clear_all_bp();
    }
    else
    {
        err_code[1] = SYNTAX_ERROR;
    }
    /* show info */
    uart_puts(err_code);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: clear_all_bp
|*
|*  PARAMETERS: void
|*
|*  DESCRIPTION: clear internal break point tables
|*
|*  RETURNS: ERR_NOTBREAK if no breakpoint was set,
|*           ERR_NORMAL if any breakpoint was cleared   
|*
 */
char clear_all_bp(void)
{
    char ret;

    ret = ERR_NOTBREAK;
    if  (!(break_points[0] & 0x80000000))
    {
        break_points[0] = 0x80000000;
        ret = ERR_NORMAL;
    }
    if (!(break_points[1] & 0x80000000))
    {
        break_points[1] = 0x80000000;
        ret = ERR_NORMAL;
    }
#ifdef HAS_4BREAKPOINTS
    if (!(break_points[2] & 0x80000000))
    {
        break_points[2] = 0x80000000;
        ret = ERR_NORMAL;
    }
    if (!(break_points[3] & 0x80000000))
    {
        break_points[3] = 0x80000000;
        ret = ERR_NORMAL;
    }
#endif /* HAS_4BREAKPOINTS */
    return ret;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_gb
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: Execute program with breakpoints set (if available).
|*               command (CMD_GB): Start from current PC-address.
|*               command (CMD_GB)(Starting address): Start from starting address.
|*
|*  RETURNS: STATE_RESET (command finished or error occured)
|*           STATE_CMD (so far, so good...)
|*
 */
void cmd_gb( void )
{
    unsigned char go_addr = 0;

    /* check if user program is executing */
    if ( application_running )
    {
        uart_puts("(4)");
        return;
    }
    if ( cmd_getval56(6) != CTRL_END )
    {
        go_addr = 1;
        cmd_getendchar();
    }
    uart_puts(err_code);
    uart_putchar(CTRL_END);
    if ( err_code[1] == ERR_NORMAL )
    {
        if (go_addr)
        {
            m16c_set_PC( val56 );
        }
        step_or_go = GO;
        application_running = 1;
    }
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_prog
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: answer the "execution state" of the user program
|*
|*  RETURNS: STATE_RESET
|*  
 */
void cmd_prog( void )
{
    /* incoming byte should be CTRL_END */
    cmd_getendchar();
    /* show info */
    uart_puts(err_code);
    /* check if user program is running or not, resp. ret '0' and '1' */
    uart_putchar( application_running ? '0' : '1' );
    uart_putchar( CTRL_END );
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_stop
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: stop user program
|*
|*  RETURNS: STATE_RESET
|*
 */
void cmd_stop( void )
{
    /* incoming byte should be CTRL_END */
    cmd_getendchar();
    /* show info */
    uart_puts(err_code);
    uart_putchar('1');
    m16c_dump_reg(0x11); /* get PC */
    uart_putchar(CTRL_END);

    application_running = 0;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_step
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: step n times
|*
|*  RETURNS:  STATE_RESET if done or error
|*              STATE_CMD, so far so good...
|*
 */
void cmd_step( void )
{
    /* check if user program is executing */
    if ( application_running )
    {
        uart_puts("(4)");
    }
    else
    {
        val4 = 1;
        if ( cmd_getval4() != CTRL_END )
        {
            cmd_getendchar();
        }
    }
    if ((err_code[1]==ERR_NORMAL) && (val4>0))
    {
        step_or_go = STEP;

        err_code[1] = m16c_step(&val4);
    }
    uart_puts(err_code);
    uart_putchar(hex[ (val4 >> 12) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  8) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  4) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  0) & 0x0000000F ]);
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_xreg
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: get or set register(s)
|*
|*  RETURNS:  STATE_RESET if done or error
|*            STATE_CMD, so far so good...
 */
void cmd_xreg( void )
{
    enum { SHOWALL, SHOWREG, SETREG };
    unsigned char action = SHOWALL;
    unsigned char regnr = 0;

    /* check if user program is executing */
    if ( application_running )
    {
        err_code[1] = ERR_EXEC;
    }
    else if ( cmd_getval2() == CTRL_END )
    {
        /* (40) = show all registers */
        action = SHOWALL;
    }
    else if ( cmd_getval4() == CTRL_END )
    {
        /* (40ii) show register ii */
        action = SHOWREG;
    }
    else
    {
        regnr = val2;
        /* (40iivvvv[vv]) set register ii */
        if ( (regnr == REG_PC) || (regnr == REG_INTB) )
        {
            cmd_getval2();
            val56 = ((unsigned long)val4 << 8) | val2;
        }
        else
        {
            val56 = val4;
        }
        if ( uart_getchar() != CTRL_END )
        {
            err_code[1] = SYNTAX_ERROR;
        }
        action = SETREG;
    }
    uart_puts(err_code);
    if ( err_code[1] == ERR_NORMAL )
    {
        switch (action)
        {
        case SHOWALL:
            for ( regnr = 0; regnr < REG_MAX; ++regnr )
            {
                m16c_dump_reg(regnr);
            }
            break;
        case SHOWREG:
            m16c_dump_reg(val2);
            break;
        default:
            m16c_set_reg(regnr, val56);
            break;
        }
    }
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_dump
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: dump part of memory 
|*
|*  RETURNS:  STATE_RESET if done or error
|*            STATE_CMD, so far so good...
|*
 */
void cmd_dump( void )
{
    unsigned char type;

    val4 = 1;  /* set default number of bytes to dump */
    type = cmd_getnibble();
    if ( cmd_getval56(5) != CTRL_END )
    {
        /* this is not (41t), that means that the
         * first nibble we just read ('type') is
         * actually the high byte of the address
         */
        dump_address = val56 | ((unsigned long)type << 5*4);
        type = cmd_getnibble();
        if ( (cmd_getval4() != CTRL_END) && (uart_getchar() != CTRL_END) )
        {
            err_code[1] = SYNTAX_ERROR;
        }
    }
    uart_puts( err_code );
    if ( err_code[1] == ERR_NORMAL )
    {
        m16c_dump( val4, type );
    }
    uart_putchar(CTRL_END);
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_fill
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: fill part of memory 
|*
|*  RETURNS:  STATE_RESET if done or error
|*              STATE_CMD, so far so good...
|*
 */
void cmd_fill( void )
{
    unsigned int n;
    unsigned char type;

    cmd_getval56(6);
    cmd_getval4();
    n = val4;
    type = uart_getchar();
    cmd_getval4();
    if ( uart_getchar() == CTRL_END )
    {
        uart_puts( err_code );
        if ( err_code[1] == ERR_NORMAL )
        {
            m16c_fill( val56, n, type, val4 );
        }
    }
    else
    {
        err_code[1] = SYNTAX_ERROR;
        uart_puts( err_code );
    }
    uart_putchar( CTRL_END );
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_fm
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: 
|*
|*  RETURNS: 
|*
 */
void cmd_fm( void )
{
    uart_getchar();
    uart_getchar();
    uart_puts( err_code );
    uart_putchar( CTRL_END );
}

void cmd_debug( void )
{
    unsigned int i;

    cmd_getval56(6);
    cmd_getval4();
    while ( uart_getchar() != CTRL_END )
    {
        /* nop */ ;
    }
    uart_puts( err_code );
    for ( i = 0; i<val4; ++i )
    {
        unsigned char val = ((__far unsigned char *)val56)[i];
        uart_putchar( hex[(val >> 4)  & 0xf] );
        uart_putchar( hex[(val >> 0)  & 0xf] );
        uart_putchar( '-' );
    }
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_err
|*
|*  PARAMETERS: c, incoming byte for this command 
|*
|*  DESCRIPTION: no function selected yet, just return STATE_RESET
|*
|*  RETURNS:  STATE_RESET
|*
 */
void cmd_err( void )
{
    return;
}

/**********************************************************************\
|*
|*  FUNCTION: cmd_process_new_char
|*
|*  PARAMETERS: c, incoming character
|*
|*  DESCRIPTION: main level state machine that handles incoming
|*              characters from uart
|*                     
 */
void cmd_handle_command( void )
{
    while ( uart_getchar() != CTRL_START )
    {
        /* nop */ ;
    }
    setjmp( env );
    err_code[1] = ERR_NORMAL;
    /* get the command value */
    cmd_getval2();
    /* select the corresponding function */
    cmd_set_handle( val2 );
    /* and call it to handle followup characters */
    cmd_handle();
    if ( application_running )
    {
        /* use the BRK interrupt routine to (re)start the user application */
        __brk();
    }
}

