/**********************************************************************
|*
|*  Version:    1.1
|*
|*  Description:
|*
|*      UART driver for the TASKING M16C ROM monitor
|*
|*  Copyright 2003 Altium BV
|*
 **********************************************************************/

#include <stdlib.h>
#include "target.h"
#include "uart.h"
#include "m16c.h"
#include "command.h"


/**********************************************************************
|*
|*  Function:           uart_init
|*
|*  Description:        Initialize the serial I/O port
|*
 */
void uart_init( void )
{
    /* Initialize serial port */
    UxC0        = 0x00;                 // Disable serial I/O
    UxBRG       = BRG;
    UxMR        = 0x05;                 // serial mode select -> 8-bits, internal clock, one stop bit, no parity 
#ifdef HANDSHAKE_HARDWARE
    UxC0        = 0x00 | FCLOCK;        // FCLOCK (0 = f1, 1 = f8, 2 = f32) source, CTS/RTS enabled
#else
    UxC0        = 0x10 | FCLOCK;        // FCLOCK (0 = f1, 1 = f8, 2 = f32) source, CTS/RTS disabled
#endif
    SxRIC       = 0x07;                 // Set receive interrupts on, interrupt level 7
    UxC1        = 0x05;                 // Transmission/reception enabled
    __nop();
    __nop();

    __fset( __I );
}

/**********************************************************************
|*
|*  Function: uart_putchar
|*
|*  Parameters: byte = byte to be sent
|*
|*  Description: send a byte through the serial port
|*
 */
void uart_putchar( unsigned char byte )
{
    while( !UxC1_TI )
    {
        /* NOP */ ;
    }
    UxTB = byte;
}

/**********************************************************************
|*
|*  Function: uart_getchar
|*
|*  Returns: Character read o.k.: 0
|*                                1 otherwise
|*
|*  Description: Get a character from the serial port
|*
 */
unsigned char uart_getchar( void )
{
    char x;

    do
    {
        x = se_buf_in - se_buf_out;     /* calculate current size of input buffer */

        if (x < 0)
        {
            x += INPUT_BUFFER_SIZE+1;  /* wrap around */
        }
        if ((x == LOW_WATER_MARK) && flag_xoff_sent)
        {
            uart_putchar( XON );
            flag_xoff_sent = 0;
        }
    } while (!x);  /* wait until a character is available */

    x = se_buf[se_buf_out]; /* get character from serial buffer */

    se_buf_out++;       /* increment "out" pointer */
    if (se_buf_out == INPUT_BUFFER_SIZE+1)
    {
        se_buf_out = 0; /* wrap around */
    }
    return x;
}

/**********************************************************************
|*
|*  Function: uart_puts
|*
|*  Parameters: str = pointer to the string to be written, may not
|*              be a NULL pointer!
|*
|*  Description: output an ASCIIZ string to the serial port
|*
 */
void uart_puts( const char * str )
{
    while( *str )
    {
        uart_putchar( *str++ );
    }
}

/**********************************************************************
|*
|*  Function: uart_rxd_interrupt
|*
|*  Description: handle receiver interrupts from the serial port
|*
|*  Attention: this interrupt routine MUST be located at 0x0FE000,
|*             the user application has to deal with it.
|*                     
 */
static void 
__frame(R0,R1,R2,R3,A0,A1,SB,FB) __interrupt( RXDxIRQ )
uart_rxd_interrupt( void ) __at(0x0FE000)
{
    unsigned char ch;
    char size;

    /* check if we received a valid character */
    if ( UxRBH & 0x80 )
    {
        /* some communication error happened, reset error flags */
        UxC1_RE = 0;  /* error flags are reset by disabling receive enable bit */
        UxC1_RE = 1;  /* now set the receive enable bit again */
        /* invalid character, just return */
        return;
    }
    ch = UxRBL;
    if ( ch == '!' )
    {
        ta_reset();
    }
    /*
     * check if we have to switch to to monitor's command handler
     */
    if ( user_mode == 1 )
    {
        /* Apparently the user application is running. Now
         * only switch to the monitor command handler if the
         * character received is the command start character.
         * Any other characters are just ignored.
         */
        if ( ch == CTRL_START )
        {
            switch_to_monitor_mode();
        }
        else
        {
            /* ignore this character */
            return;
        }
    }
    /* store character in input buffer */
    se_buf[se_buf_in] = ch;
    /* increment "in" pointer */
    if (++se_buf_in == INPUT_BUFFER_SIZE + 1)
    {
        se_buf_in = 0;          /* wrap around */
    }
    size = se_buf_in - se_buf_out;
    if (size < 0)
    {
        size += INPUT_BUFFER_SIZE+1;  /* wrap around */
    }
    if (!size) /* check for buffer overflow (se_buf_in == se_buf_out) */
    {
        /* buffer overflow! */
        se_buf_in--;
        if (se_buf_in < 0)
        {
            se_buf_in = INPUT_BUFFER_SIZE+1;  /* wrap around */
        }
    }
    else if (size == HIGH_WATER_MARK) /* size can be negative.... */
    {
        /* high water mark reached, stop sender */
        uart_putchar( XOFF );
        flag_xoff_sent = 1;     /* remember that we sent XOFF */
    }
}
